@@ -1,9 +1,9 @@
 # -*- coding: utf-8 -*-
 #
-# This file is part of the instrument simulator project
+# This file is part of the sinstruments project
 #
-# Copyright (c) 2018 Tiago Coutinho
-# Distributed under the MIT. See LICENSE for more info.
+# Copyright (c) 2018-present Tiago Coutinho
+# Distributed under the GPLv3 license. See LICENSE for more info.
 
 """Instrument Simulator
 
@@ -14,6 +14,7 @@ To start the server you can do something like::
 
 from __future__ import print_function
 
+import io
 import os
 import pty
 import sys
@@ -178,7 +179,7 @@ class SerialServer(SimulatorServerMixin):
     """
 
     def __init__(self, name, handler, **kwargs):
-        self.address = kwargs.pop("url")
+        self.address = kwargs.pop("url", '')
         self.set_listener(kwargs.pop('listener', None))
         SimulatorServerMixin.__init__(self, name, handler, **kwargs)
 
@@ -222,7 +223,7 @@ class SerialServer(SimulatorServerMixin):
         self.handle(self.fileobj)
 
     def broadcast(self, msg):
-        self.fileobj.write(msg)
+        self.send(self.fileobj, msg)
 
     def send(self, channel, data):
         delay(len(data), baudrate=self.baudrate)
@@ -246,10 +247,17 @@ class TCPServer(StreamServer, SimulatorServerMixin):
         info = self._log.info
         info("new connection from %s", addr)
         channel = sock.makefile('rwb', 0)
+        # non buffered rwb are SocketIO objects (not instances of io.BufferedIOBase)
+        # so they don't have read1.
+        def read1(size=-1):
+            size = io.DEFAULT_BUFFER_SIZE if size == -1 else size
+            return channel.read(size)
+        channel.read1 = read1
         self.connections[addr] = sock
         try:
             SimulatorServerMixin.handle(self, channel)
         finally:
+            del channel
             del self.connections[addr]
             sock.close()
         info("client disconnected %s", addr)
@@ -303,8 +311,9 @@ class BaseDevice(object):
     newline = b"\n"
     baudrate = None
 
-    def __init__(self, name, **kwargs):
+    def __init__(self, name, server=None, **kwargs):
         self.name = name
+        self.server = server
         self.newline = kwargs.pop('newline', self.newline)
         self._log = logging.getLogger("{0}.{1}".format(_log.name, name))
         self.__transports = weakref.WeakKeyDictionary()
@@ -353,7 +362,7 @@ class Server(object):
     Handles a set of devices
     """
 
-    def __init__(self, devices=(), backdoor=None):
+    def __init__(self, devices=(), backdoor=None, registry=None):
         self._log = _log
         self._log.info("Bootstraping server")
         if backdoor:
@@ -372,6 +381,7 @@ class Server(object):
         else:
             self._log.info("no backdoor declared")
 
+        self.registry = {} if registry is None else registry
         self.devices = {}
         for device in devices:
             try:
@@ -385,27 +395,28 @@ class Server(object):
 
     def create_device(self, device_info):
         klass_name = device_info.get("class")
-        name = device_info.get("name", klass_name)
+        name = device_info["name"]
+        # name should be unique
+        assert name not in self.devices
         self._log.info("Creating device %s (%r)", name, klass_name)
-        device, transports = create_device(device_info)
-        self.devices[device] = transports
+        device_info["server"] = self
+        device, transports = create_device(device_info, self.registry)
+        self.devices[name] = device
         return device, transports
 
     def get_device_by_name(self, name):
-        for device in self.devices:
-            if device.name == name:
-                return device
+        return self.devices[name]
 
     def start(self):
         tasks = []
-        for device in self.devices:
-            for interface in self.devices[device]:
-                tasks.append(gevent.spawn(interface.serve_forever))
+        for device in self.devices.values():
+            for transport in device.transports:
+                tasks.append(gevent.spawn(transport.serve_forever))
         return tasks
 
     def stop(self):
-        for device in self.devices:
-            for transport in self.devices[device]:
+        for device in self.devices.values():
+            for transport in device.transports:
                 transport.stop()
 
     def serve_forever(self):
@@ -419,19 +430,21 @@ class Server(object):
         return "{0}({1})".format(self.__class__.__name__, self.name)
 
 
-def create_device(device_info):
+def create_device(device_info, registry):
     device_info = dict(device_info)
     class_name = device_info.pop("class")
-    module_name = device_info.pop("module", class_name.lower())
+    module_name = device_info.pop("module", None)
     package_name = device_info.pop("package", None)
-    name = device_info.pop("name", class_name)
+    name = device_info.pop("name")
 
-    if package_name is None:
-        package_name = "sinstruments.simulators." + module_name
-
-    __import__(package_name)
-    package = sys.modules[package_name]
-    klass = getattr(package, class_name)
+    if package_name is None and module_name is None:
+        klass = registry[class_name].load()
+    else:
+        if package_name is None:
+            package_name = "sinstruments.simulators." + module_name
+        __import__(package_name)
+        package = sys.modules[package_name]
+        klass = getattr(package, class_name)
     device = klass(name, **device_info)
 
     transports_info = device_info.pop("transports", ())
@@ -451,6 +464,18 @@ def create_device(device_info):
     return device, transports
 
 
+def load_device_registry():
+    """
+    Return device classes for those devices which registered themselves with an
+    entry point.
+    """
+    import pkg_resources
+    return {
+        ep.name: ep
+        for ep in pkg_resources.iter_entry_points('sinstruments.device')
+    }
+
+
 def parse_config_file(file_name):
     ext = os.path.splitext(file_name)[-1]
     if ext.endswith('toml'):
@@ -469,7 +494,8 @@ def parse_config_file(file_name):
 
 def create_server_from_config(config):
     backdoor, devices = config.get("backdoor", None), config.get("devices", ())
-    return Server(devices=devices, backdoor=backdoor)
+    registry = load_device_registry()
+    return Server(devices=devices, backdoor=backdoor, registry=registry)
 
 
 def main():
